module SimplexIOInterface(
  initSimplex,
  showState,
  assert,
  check,
  solution,
  checkpoint,
  backtrack,
  Relation(..),
  SimplexInput,
  Index,
  Variable,
  exampleIOInvocation,
  exampleSlidesInput,
  exampleRun) 
  where

import SimplexInternals hiding (initSimplex)
import qualified SimplexInternals as S
import SimplexCommon
import System.IO.Unsafe
import Data.IORef

data SimplexIO = Uninit | SI Bool Integer [Integer] [(String,([Integer],SimplexCheckpoint))] SimplexState

globalState :: IORef SimplexIO
globalState = unsafePerformIO (newIORef Uninit)

showState :: IO ()
showState = do
  ss <- readIORef globalState
  case ss of
     Uninit -> putStrLn "<uninitialized>"
     SI True i is _ _ -> putStrLn $ "Okay <state " ++ show i ++ ", asserted: " ++ show is ++ ">" 
     SI False _ _ cs _ -> putStrLn $ "Unsat detected, use backtrack to one of " ++ show (map fst cs) 

initSimplex :: SimplexInput -> IO ()
initSimplex ics = do
  let ss = S.initSimplex (convertInput ics)
  ss `seq` writeIORef globalState (SI True 0 [] [] ss)
  showState

assert :: Integer -> IO ()
assert idx = do
  ss <- readIORef globalState
  case ss of
     Uninit -> error "simplex state not initialized, invoke initS first"
     SI False _ _ cs _ -> error $ "Unsat detected, use backtrack to one of " ++ show (map fst cs) 
     SI True i is cs ss ->  case assertSimplex idx ss of
       UNSAT core -> do
         ss `seq` writeIORef globalState (SI False i (idx : is) cs ss)
         putStrLn $ "unsat-core " ++ show core ++ " detected, use backtrack to one of " ++ show (map fst cs)
       Okay ss -> let j = i + 1 in do
         ss `seq` writeIORef globalState (SI True j (idx : is) cs ss)
         showState

check :: IO ()
check = do
  ss <- readIORef globalState
  case ss of
     Uninit -> error "simplex state not initialized, invoke initS first"
     SI False _ _ cs _ -> error $ "Unsat detected, use backtrack to one of " ++ show (map fst cs) 
     SI True i is cs ss ->  case checkSimplex ss of
       (Just core, ss) -> do
         ss `seq` writeIORef globalState (SI False i is cs ss)
         putStrLn $ "unsat-core " ++ show core ++ " detected, use backtrack to one of " ++ show (map fst cs)
       (Nothing, ss) -> let j = i + 1 in do
         ss `seq` writeIORef globalState (SI True j is cs ss)
         showState

solution :: IO ()
solution = do
  ss <- readIORef globalState
  case ss of
     Uninit -> error "simplex state not initialized, invoke initS first"
     SI False _ _ cs _ -> error $ "Unsat detected, use backtrack to one of " ++ show (map fst cs) 
     SI True _ _ _ ss -> do
       putStrLn $ show (solutionSimplex ss)

checkpoint :: IO ()
checkpoint = do
  ss <- readIORef globalState
  case ss of
     Uninit -> error "simplex state not initialized, invoke initS first"
     SI False _ _ cs _ -> error $ "Unsat detected, use backtrack to one of " ++ show (map fst cs) 
     SI True i idx cs ss -> let
        cp = checkpointSimplex ss
        j = i + 1
        cj = "cp" ++ show j
        new = SI True j idx ((cj, (idx,cp)) : cs) ss
        in do
             cp `seq` writeIORef globalState new
             putStrLn $ "checkpoint \"" ++ cj ++ "\" created"
             showState
       
backtrack :: String -> IO ()
backtrack c = do
  ss <- readIORef globalState
  case ss of
     Uninit -> error "simplex state not initialized, invoke initS first"
     SI _ i _ cs ss -> case lookup c cs of
        Nothing -> error $ "checkpoint " ++ c ++ " unknown, known checkpoints: " ++ show (map fst cs) 
        Just (idx,cp) -> let
           ssn = backtrackSimplex cp ss
           new = SI True (i+1) idx (dropWhile ((/= c) . fst) cs) ssn
           in do 
              ssn `seq` writeIORef globalState new
              showState
              
exampleIOInvocation = putStrLn $ unlines [
  "initSimplex [" ++
   "(1, [(2/3 :: Prelude.Rational, \"x\"), (-3 / 2, \"y\")], Gt, 4), " ++
   "(2, [(1, \"y\")], Lt, -4), " ++
   "(3, [(1, \"y\")], Ge, 2)]",
  "assert 1",
  "assert 2",
  "check",
  "solution",
  "checkpoint",
  "assert 3",
  "backtrack \"cp4\""]

exampleSlidesInput :: SimplexInput
exampleSlidesInput = [
  (1, [(1, "x")], Ge, 5),
  (-1, [(1, "x")], Lt, 5),
  (2, [(2, "x"), (1, "y")], Le, 12),
  (-2, [(2, "x"), (1, "y")], Gt, 12),
  (3, [(1, "y")], Lt, 3), 
  (-3, [(1, "y")], Ge, 3), 
  (4, [(1, "x"), (-3, "y")], Le, 2),
  (-4, [(1, "x"), (-3, "y")], Gt, 2)
 ]

exampleRun = do
  initSimplex exampleSlidesInput
  assert 1
  assert (-3)
  assert 4
  check -- sat
  checkpoint -- delivers checkpoint "cp5"
  assert 2
  check -- unsat core [1,2,-3]
  backtrack "cp5" -- back at asserting [1,-3,4]

  
